package processors;

import haxe.ds.IntMap;
import clay.Entity;
import clay.Vector;
import clay.Text;
import clay.Processor;
import clay.NodeList;
import physics.PointMass;
import clay.structural.Pool;
import clay.utils.Maths;

import utils.ShapeDrawer;

import physics.Contact;
import physics.collision.Intersect;


class BroadPhaseProcessor extends Processor {


	public var contactsCount(default, null):Int;

	public var pos:Vector;

		/* the square cell gridLength of the grid. Must be larger than the largest shape in the space. */
	public var cellSize(default, set):Vector;

	public var min(default, null):Vector;
	public var max(default, null):Vector;

	public var width(default, null):Float;
	public var height(default, null):Float;

		/* the world space width */
	public var w (default, null):Int;
		/* the world space height */
	public var h (default, null):Int;

		/* the number of buckets (i.e. cells) in the spatial grid */
	public var gridLength (default, null):Int;
		/* the array-list holding the spatial grid buckets */
	public var grid(default, null) : haxe.ds.Vector<Array<PointMass>>;

	// var powerOfTwo:UInt;

	var vectorPool:Pool<Vector>;
	var contactsPool:Pool<Contact>;

	// temp objects
	var _tmp_getGridIndexesArray:Array<Int>;
	var _tmp_getEntitiesArray:Array<Entity>;
	var _tmp_getCollisionObjectsArray:Array<PointMass>;

	#if debugSpace

		var cText:Text;

		var debugBoxes : haxe.ds.Vector<DebugObject>;

	#end


	public function new(_min:Vector = null, _max:Vector = null, _cellSize:Vector) {

		super();

		contactsCount = 0;

		pos = new Vector();

		min = new Vector();
		max = new Vector(1280, 720);

		if(_min != null){
			min.copy_from(_min);
		}

		if(_max != null){
			max.copy_from(_max);
		}

		width = max.x - min.x;
		height = max.y - min.y;

		cellSize = _cellSize;

		w = Math.ceil(width / cellSize.x);
		h = Math.ceil(height / cellSize.y);

		gridLength = Std.int(w * h);

		grid = new haxe.ds.Vector(gridLength);
		for (i in 0...gridLength) {
			grid[i] = new Array<PointMass>();
		}

		// temp 
		vectorPool = new Pool(10, function() {return new Vector();});
		contactsPool = new Pool( 100, function() { return new Contact(); });

		_tmp_getGridIndexesArray = [];
		_tmp_getEntitiesArray = [];
		_tmp_getCollisionObjectsArray = [];

		#if debugSpace

			cText = new Text({
				name : 'contactsText',
				pos : new Vector(8, 8),
				point_size : 14,
				text : 'contacts'
			});

			debugBoxes = new haxe.ds.Vector(gridLength);

			var iw:Int = 0;
			var ih:Int = 0;
			for (i in 0...gridLength) {

				debugBoxes[i] = new DebugObject(new Vector(pos.x + min.x + (cellSize.x * 0.5) + cellSize.x * iw, pos.x + min.y + (cellSize.y * 0.5) + cellSize.y * ih), cellSize);

				iw++;
				if(iw >= w){
					iw = 0;
					ih++;
				}
			}

		#end

	}


// Collision api

	public function getLineGridPointMasses(start:Vector, end:Vector, ?into:Array<PointMass>, ?layerMask:Int = -1):Array<PointMass> {

		var _gridIndexes:Array<Int> = getLineGridIndexes(start.x, start.y, end.x, end.y);
		var cArr:Array<PointMass> = getContactsFromIdx(_gridIndexes, layerMask);
		
		#if debugSpatialHash 

			drawGridIndexes(_gridIndexes);
			ShapeDrawer.drawLine(start, end, ShapeDrawer.color_white);

		#end

		if(cArr.length > 0){
			if (into == null) {
				into = [];
			}

			for (c in cArr) {
				if(into.indexOf(c) == -1) {
					into.push(c);
				}
			}	

			if(into.length > 0){
				return into;
			} else {
				return null;
			}
		}

		return null;

	}

// Area
	public function getAreaOverlap(_min:Vector, _max:Vector, ?layerMask:Int = -1):Bool {

		var _gridIndexes:Array<Int> = getAreaGridIndexes(_min, _max);
		var cArr:Array<PointMass> = getContactsFromIdx(_gridIndexes, layerMask);
		
		#if debugSpatialHash 

			drawGridIndexes(_gridIndexes);
			ShapeDrawer.drawBox_MinMax(_min, _max, ShapeDrawer.color_blue);

		#end

		if(cArr.length > 0){

			var aHalf:Vector = vectorPool.get();
			var aCenter:Vector = vectorPool.get();

			aCenter.x = (_min.x + _max.x) * 0.5;
			aCenter.y = (_min.y + _max.y) * 0.5;

			aHalf.x = _max.x - aCenter.x;
			aHalf.y = _max.y - aCenter.y;


			for (c in cArr) {
				if(Intersect.PointAABB(c.position, aCenter, aHalf)){

					vectorPool.put(aCenter);
					vectorPool.put(aHalf);

					return true;
				}
			}
		}

		return false;

	}

	public function getAreaOverlapNearest(_min:Vector, _max:Vector, ?layerMask:Int = -1):PointMass {

		var _gridIndexes:Array<Int> = getAreaGridIndexes(_min, _max);
		var cArr:Array<PointMass> = getContactsFromIdx(_gridIndexes, layerMask);
		
		#if debugSpatialHash 

			drawGridIndexes(_gridIndexes);
			ShapeDrawer.drawBox_MinMax(_min, _max, ShapeDrawer.color_blue);

		#end

		if(cArr.length > 0){

			var _tmpDistVec:Vector = vectorPool.get();
			var aHalf:Vector = vectorPool.get();
			var aCenter:Vector = vectorPool.get();

			aCenter.x = (_min.x + _max.x) * 0.5;
			aCenter.y = (_min.y + _max.y) * 0.5;

			aHalf.x = _max.x - aCenter.x;
			aHalf.y = _max.y - aCenter.y;

			var closest:PointMass = null;

			var minDist:Float = Math.POSITIVE_INFINITY;
			var _tmpDist:Float = 0;

			for (c in cArr) {

				if(Intersect.PointAABB(c.position, aCenter, aHalf)){

					_tmpDistVec.x = aCenter.x - c.position.x;
					_tmpDistVec.y = aCenter.y - c.position.y;

					_tmpDist = _tmpDistVec.lengthsq;

					if( _tmpDist < minDist ){
						minDist = _tmpDist;
						closest = c;
					}

				}
			}

			vectorPool.put(_tmpDistVec);
			vectorPool.put(aHalf);
			vectorPool.put(aCenter);

			return closest;
		}

		return null;

	}

	public function getAreaOverlapAll(_min:Vector, _max:Vector, ?into:Array<PointMass>, ?layerMask:Int = -1):Array<PointMass> {

		var _gridIndexes:Array<Int> = getAreaGridIndexes(_min, _max);
		var cArr:Array<PointMass> = getContactsFromIdx(_gridIndexes, layerMask);
		
		#if debugSpatialHash 

			drawGridIndexes(_gridIndexes);
			ShapeDrawer.drawBox_MinMax(_min, _max, ShapeDrawer.color_blue);

		#end

		if(cArr.length > 0){

			if (into == null) {
				into = [];
			}

			var aHalf:Vector = vectorPool.get();
			var aCenter:Vector = vectorPool.get();

			aCenter.x = (_min.x + _max.x) * 0.5;
			aCenter.y = (_min.y + _max.y) * 0.5;

			aHalf.x = _max.x - aCenter.x;
			aHalf.y = _max.y - aCenter.y;


			for (c in cArr) {

				if(Intersect.PointAABB(c.position, aCenter, aHalf)){
					into.push(c);
				}
			}

			vectorPool.put(aHalf);
			vectorPool.put(aCenter);

			if(into.length > 0){
				return into;
			} else {
				return null;
			}
		}

		return null;

	}

	override function onAdded() {

		#if debugSpace
			Clay.on(clay.Ev.render, draw);
		#end

	}

	override function onRemoved() {

		#if debugSpace
			Clay.off(clay.Ev.render, draw);
		#end

	}

	// @:allow(processors.PhysicsProcessor) 
	@:allow(physics.PointMass) 
	inline function addToGrid(p:PointMass){
		
		// collider.broadPhase = this;

		// collider.updateBounds = updateBounds;
		// collider.updateBroadphase = updateBroadphase;

		// updateBounds(p);
		updateBroadphase(p);

	}

	// @:allow(processors.PhysicsProcessor) 
	@:allow(physics.PointMass) 
	inline function removeFromGrid(p:PointMass){
		
		removeIndexes(p);

		// collider.updateBounds = null;
		// collider.updateBroadphase = null;

		// collider.broadPhase = null;

	}

	public inline function updateBroadphase(p:PointMass){

		// trace('updateBroadphase');

		var pLast:PointMass = p.prev;
		var pNext:PointMass = p.next;

		var ar:Array<Int> = [];
		if(pLast != null) {
			var ab:Array<Int> = getLineGridIndexes(pLast.position.x, pLast.position.y, p.position.x, p.position.y);
			for (i in ab) {
				ar.push(i);
			}
		}
		if(pNext != null) {
			var bc:Array<Int> = getLineGridIndexes(p.position.x, p.position.y, pNext.position.x, pNext.position.y);
			for (i in bc) {
				ar.push(i);
			}
		}

		updateIndexes(p, ar);

	}

	// @:allow(processors.PhysicsProcessor) 
	// inline function removeContact(id:Int, contacts:IntMap<Contact>) {

	// 	var c:Contact = contacts.get(id);

	// 	if(c != null){
	// 		contacts.remove(id);
	// 		contactsCount--;
			
	// 		c.clear();
	// 		contactsPool.put(c);
	// 	}

	// }

	// inline function addContactTo(coll:PointMass, otherColl:PointMass, into:IntMap<Contact>) {
		
	// 	var id:Int = coll.id + otherColl.id;

	// 	var c:Contact = into.get(id);

	// 	if(c == null){
	// 		c = contactsPool.get();
	// 		c.id = id;
	// 		c.collider = coll;
	// 		c.otherCollider = otherColl;
	// 		into.set(id,c);
	// 		contactsCount++;
	// 	} else {
	// 		c.remove = false;
	// 	}

	// }

	inline function checkContacts(contacts:IntMap<Contact>) {

		for (c in contacts) {
			if(c.remove){
				// removeContact(c.id, contacts);
			} else {
				c.remove = true;
			}
		}

	}

	public inline function getContacts(coll:PointMass, _type:Int = -1, into:IntMap<Contact> = null):IntMap<Contact> {

		if(into == null){
			into = new IntMap();
		}

		// проверяем все яйчейки
		for (i in coll.gridIndex) { 
			// проверяем кто есть в ней
			for (otherColl in grid[i]) {

				if(coll == otherColl) {
					continue;
				}

				// фильтруем по типу ?
				if(_type != -1 && _type != otherColl.type) {
					continue;
				}

				// addContactTo(coll, otherColl, into);

			}
		}

		checkContacts(into);

		return into;

	}

// internal api

	inline function getContactsFromIdx(gridIndexes:Array<Int>, layerMask:Int):Array<PointMass> {

		var ret:Array<PointMass> = _tmp_getCollisionObjectsArray;

		if(ret.length != 0) {
			ret.splice(0, ret.length);
		}
		
		for (i in gridIndexes) {
			for (p in grid[i]) { // проверка на grid[i] == null ?
				if(p != null){
					ret.push(p);
				}
			}
		}

		return ret;

	}

	inline function getAreaGridIndexes(_min:Vector, _max:Vector):Array<Int> {

		var ret:Array<Int> = _tmp_getGridIndexesArray;

		if(ret.length != 0) {
			ret.splice(0, ret.length);
		}

		if(!overlaps(_min, _max)) {
			return ret;
		}

		if(_min.x == _max.x && _min.y == _max.y){
			ret[0] = getIndex1DFromVec(_min);
		} else {
			var aabbMinX:Int = Maths.clampi(getIndex_X(_min.x), 0, w-1);
			var aabbMinY:Int = Maths.clampi(getIndex_Y(_min.y), 0, h-1);
			var aabbMaxX:Int = Maths.clampi(getIndex_X(_max.x), 0, w-1);
			var aabbMaxY:Int = Maths.clampi(getIndex_Y(_max.y), 0, h-1);

			var aabbMin:Int = getIndex1D(aabbMinX, aabbMinY);
			var aabbMax:Int = getIndex1D(aabbMaxX, aabbMaxY);

			ret.push(aabbMin);

			if(aabbMin != aabbMax) {
				ret.push(aabbMax);

				var lenX:Int = aabbMaxX - aabbMinX + 1;
				var lenY:Int = aabbMaxY - aabbMinY + 1;

				for (x in 0...lenX) {
					for (y in 0...lenY) {
						// пропускаем добавленые
						if((x == 0 && y == 0) || (x == lenX-1 && y == lenY-1) ){
							continue;
						}

						ret.push(getIndex1D(x, y) + aabbMin);
					}
				}
			}
		}
		
		return ret;

	}

	function getLineGridIndexes(x1:Float, y1:Float, x2:Float, y2:Float):Array<Int>{

		var ret:Array<Int> = _tmp_getGridIndexesArray;

		if(ret.length != 0) {
			ret.splice(0, ret.length);
		}

		// convert position to grid coordinates
		x1 = getGridPos_X(x1);
		y1 = getGridPos_Y(y1);
		x2 = getGridPos_X(x2);
		y2 = getGridPos_Y(y2);

		var gridPosX:Int = getIndexNoPos_X(x1);
		var gridPosY:Int = getIndexNoPos_Y(y1);

		if(!isValidGridIdx2D(gridPosX, gridPosY)) {
			return ret;
		}
		
		ret.push(getIndex1D(gridPosX, gridPosY));

		var dirX:Float = x2 - x1;
		var dirY:Float = y2 - y1;
		var distSqr:Float = dirX * dirX + dirY * dirY;

		if (distSqr < 1.0E-8) {
			return ret;
		}
		
		var nf:Float = 1 / Math.sqrt(distSqr);
		dirX *= nf;
		dirY *= nf;

		var deltaX:Float = cellSize.x / Math.abs(dirX);
		var deltaY:Float = cellSize.y / Math.abs(dirY);

		var maxX:Float = gridPosX * cellSize.x - x1;
		var maxY:Float = gridPosY * cellSize.y - y1;

		if (dirX >= 0) {
			maxX += cellSize.x;
		}

		if (dirY >= 0) {
			maxY += cellSize.y;
		}

		maxX /= dirX;
		maxY /= dirY;

		var stepX:Int = Maths.sign(dirX);
		var stepY:Int = Maths.sign(dirY);

		var gridGoalX:Int = getIndexNoPos_X(x2);
		var gridGoalY:Int = getIndexNoPos_Y(y2);

		var currentDirX:Int = gridGoalX - gridPosX;
		var currentDirY:Int = gridGoalY - gridPosY;

		while (currentDirX * stepX > 0 || currentDirY * stepY > 0) {
			if (maxX < maxY) {
				maxX += deltaX;
				gridPosX += stepX;
				currentDirX = gridGoalX - gridPosX;
			} else {
				maxY += deltaY;
				gridPosY += stepY;
				currentDirY = gridGoalY - gridPosY;
			}

			if(!isValidGridIdx2D(gridPosX, gridPosY)) {
				break;
			}
			
			ret.push(getIndex1D(gridPosX, gridPosY));
		}

		return ret;

	}	

	inline function addIndexes(p:PointMass, _cellPos:Int) {
		// trace('_cellPos: $_cellPos');
		if(grid[_cellPos].indexOf(p) == -1) {
			grid[_cellPos].push(p);
			p.gridIndex.push(_cellPos);
		}

	}

	inline function removeIndex(p:PointMass, _pos:Int) {

		#if debugSpace 
			resetCellDebug(_pos);
		#end
		
		grid[_pos].remove(p);

	}

	inline function updateIndexes(p:PointMass, _ar:Array<Int>) {

		for (i in p.gridIndex) {
			removeIndex(p, i);
		}

		p.gridIndex.splice(0, p.gridIndex.length);

		for (i in _ar) {
			addIndexes(p, i);
		}

	}

	inline function removeIndexes(p:PointMass){

		for (i in p.gridIndex) {
			removeIndex(p, i);
		}

		p.gridIndex.splice(0, p.gridIndex.length);

	}

// 1D grid
	inline function getIndex1DFromVec(_pos:Vector):Int { // i = x + w * y;  x = i % w; y = i / w;

		return Std.int((Std.int(_pos.x - (pos.x + min.x)) / cellSize.x) + w * (Std.int(_pos.y - (pos.y + min.y)) / cellSize.y));

	}

	inline function getGridPos_X(_pos:Float):Float {

		return _pos - (pos.x + min.x);

	}

	inline function getGridPos_Y(_pos:Float):Float {

		return _pos - (pos.y + min.y);

	}

	inline function toGridPos(_pos:Vector) {

		_pos.x -= pos.y + min.y;
		_pos.y -= pos.y + min.y;

	}

/*	inline function getIndex(_pos:Float):Int {

		return Std.int(_pos) >> powerOfTwo;

	}*/

	inline function getIndex_X(_pos:Float):Int {

		return Std.int((_pos - (pos.x + min.x)) / cellSize.x);

	}

	inline function getIndex_Y(_pos:Float):Int {

		return Std.int((_pos - (pos.y + min.y)) / cellSize.y);

	}

	inline function getIndexNoPos_X(_pos:Float):Int {

		return Std.int(_pos / cellSize.x);

	}

	inline function getIndexNoPos_Y(_pos:Float):Int {

		return Std.int(_pos / cellSize.y);

	}

	inline function getIndex1D(_x:Int, _y:Int):Int { // i = x + w * y;  x = i % w; y = i / w;

		return Std.int(_x + w * _y);

	}

	inline function getVecFromIdx1D(_i:Int):Vector { // i = x + w * y;  x = i % w; y = i / w;

		return new Vector(pos.x + min.x + (Std.int(_i % w) * cellSize.x + cellSize.x * 0.5), pos.y + min.y + (Std.int(_i / w) * cellSize.y + cellSize.y * 0.5));

	}

	inline function isValidGridIdx2D(_x:Int, _y:Int):Bool {

		if(_x < 0 || _x >= w) return false;
		if(_y < 0 || _y >= h) return false;

		return true;

	}

	inline function isValidGridIdx_X(num:Int):Bool {

		if(num < 0 || num >= w) return false;

		return true;

	}

	inline function isValidGridIdx_Y(num:Int):Bool {

		if(num < 0 || num >= h) return false;

		return true;

	}

	public inline function pointInside(_pos:Vector):Bool {

		if ( _pos.x < (pos.x + min.x) || _pos.x > (pos.x + max.x) ) return false;
		if ( _pos.y < (pos.y + min.y) || _pos.y > (pos.y + max.y) ) return false;

		return true;

	}

	inline function overlaps(_min:Vector, _max:Vector):Bool {

		if ( _max.x < (pos.x + min.x) || _min.x > (pos.x + max.x) ) return false;
		if ( _max.y < (pos.y + min.y) || _min.y > (pos.y + max.y) ) return false;

		return true;

	}

	public function drawGridIndexes(_indexes:Array<Int>){

		for (i in _indexes) {
			var _pos:Vector = getVecFromIdx1D(i);
			var _half:Vector = new Vector(cellSize.x * 0.5, cellSize.y * 0.5);
			ShapeDrawer.drawBox(_pos, _half);
		}

	}

	inline function set_cellSize(value:Vector):Vector {

		// powerOfTwo = clay.utils.PowerOfTwo.toPowOf2(value);
		// cellSize = clay.utils.PowerOfTwo.require(value);

		return cellSize = value;

	}


// DEBUG
	#if debugSpace 

		public function draw(_){
			
			cText.text = 'contacts: ${contactsCount}';

			ShapeDrawer.drawBox_MinMax(Vector.Add(pos, min), Vector.Add(pos, max), ShapeDrawer.color_pink);

			for (i in 0...grid.length) {
				if(grid[i].length > 0){

					// var gPos:Vector = getVecFromIdx1D(i).multiplyScalar(cellSize).addScalar(cellSize * 0.5);
					var gPos:Vector = getVecFromIdx1D(i);

					debugBoxes[i].text.transform.pos.copy_from(gPos);
					debugBoxes[i].box.transform.pos.copy_from(gPos);

					debugBoxes[i].text.visible = true;
					debugBoxes[i].text.text = "" + grid[i].length;
					debugBoxes[i].box.visible = true;

					if(grid[i].length > 1){
						debugBoxes[i].box.color = ShapeDrawer.color_orange_02;

					} else {
						debugBoxes[i].box.color = ShapeDrawer.color_gray_01;
					}

					for (c in grid[i]) {
						ShapeDrawer.drawPoint(c.position.x, c.position.y, ShapeDrawer.color_blue);
					}
					
				}
			}

		}

		function resetCellDebug(_pos:Int){

			debugBoxes[_pos].text.text = "";
			debugBoxes[_pos].box.color = ShapeDrawer.color_gray_01;

			debugBoxes[_pos].text.visible = false;
			debugBoxes[_pos].box.visible = false;

		}

	#end

}

#if debugSpace 

	private class DebugObject {

		public var text:phoenix.geometry.TextGeometry;
		public var box:phoenix.geometry.RectangleGeometry;


		public function new(_pos:Vector, _size:Vector){

			box = Clay.draw.rectangle({
				x : _pos.x,
				y : _pos.y,
				w : _size.x,
				h : _size.y,
				origin : new Vector(_size.x*0.5,_size.y*0.5),
				visible : false,
				depth : 999,
				color : ShapeDrawer.color_gray_01
			});

			text = Clay.draw.text({
				pos : _pos,
				text : "",
				visible : false,
				depth : 999,
				point_size : 12
			});

		}

	}

#end
