package processors;


import clay.Vector;
import clay.Color;
import clay.Processor;
import clay.NodeList;
import clay.Entity;
import clay.components.TransformComponent;
import clay.components.SpriteComponent;
import components.water.WaterComponent;
import components.player.SurfboardComponent;
import physics.PointMass;
import utils.ShapeDrawer;
import clay.Input;
import clay.utils.Maths;
// import physics.JVectorTools;
import phoenix.geometry.QuadGeometry;
import processors.BroadPhaseProcessor;
import physics.RayIntersection;
import physics.Line;
import clay.tween.Actuate;
import clay.tween.easing.Quad;

import nodes.WaterNode;
import nodes.SurfboardNode;
import nodes.PlayerNode;


class WaterProcessor extends Processor {


	var broadPhase:BroadPhaseProcessor;
	var waterNodes:NodeList<WaterNode>;
	var surfNodes:NodeList<SurfboardNode>;
	var playerNodes:NodeList<PlayerNode>;

	var line:Line;
	var playerTransform:TransformComponent;
	var playerSprite:SpriteComponent;

	var pointMass:PointMass;
	var water:WaterComponent;
	var surf:SurfboardComponent;
	var mousePos:Vector;
	var spawnPos:Float = 256;

	var colorA:Color;
	var colorB:Color;
	var startSpeed:Float = 512; // 448
	var speed:Float = 512;
	var stopped:Bool = false;


	public function new() {

		super();

		mousePos = new Vector();

		colorA = Colors.INDIGO.clone();
		colorA.a = 0.7;
		colorB = Colors.DEEP_KOAMARU.clone();
		colorB.a = 0.7;

	}

	override function onAdded() {

		waterNodes = world.nodes.get(WaterNode);
		if(waterNodes != null) {
			waterNodes.onadded.connect(onWaterNodeAdded);
			waterNodes.onremoved.connect(onWaterNodeRemoved);
		}

		surfNodes = world.nodes.get(SurfboardNode);
		if(surfNodes != null) {
			surfNodes.onadded.connect(onSurfNodeAdded);
			surfNodes.onremoved.connect(onSurfNodeRemoved);
		}

		playerNodes = world.nodes.get(PlayerNode);
		if(playerNodes != null) {
			playerNodes.onadded.connect(onPlayerNodeAdded);
			// playerNodes.onremoved.connect(onPlayerNodeRemoved);
		}

		broadPhase = world.processors.get(BroadPhaseProcessor);

	}

	override function onRemoved() {

		if(waterNodes != null) {
			waterNodes.onadded.disconnect(onWaterNodeAdded);
			waterNodes.onremoved.disconnect(onWaterNodeRemoved);
			waterNodes = null;
		}

		if(surfNodes != null) {
			surfNodes.onadded.disconnect(onSurfNodeAdded);
			surfNodes.onremoved.disconnect(onSurfNodeRemoved);
			surfNodes = null;
		}
		broadPhase = null;

	}

	override function onPriorityChanged(value:Int) {

		if(waterNodes != null) {
			waterNodes.onadded.updateOrder(onWaterNodeAdded, value);
			waterNodes.onremoved.updateOrder(onWaterNodeAdded, value);
		}

	}


	var broadPhaseName:String = '__broadphase';
	var pmOffset:Int = 0;
	override function update(dt:Float) {

		if(!stopped) {
			speed += 3 * dt;
		}

		if(playerTransform.pos.y < -64 && surf.velocity.y < 0) {
			surf.velocity.y = 0;
		}

		if(playerTransform.pos.y > Clay.screen.h) {
			stop();
		}

		surf.velocity.y += surf.gravity * dt;
		surf.line.y += surf.velocity.y * dt;

		surf.angularVelocity *= 1.0 - surf.angularDamping;

		if(surf.angularVelocity < - 448) {
			surf.angularVelocity = -448;
		}

		if(stopped) {
			if(surf.velocity.y < 0) {
				surf.velocity.y = 0;
			}
		}


		surf.line.x += speed * dt;

		Clay.camera.pos.x += speed * dt;
		broadPhase.pos.x = Clay.camera.pos.x;

		// ShapeDrawer.drawText(Clay.camera.pos.x + 512,256, surf.angularVelocity);

		var p1:PointMass;
		var p2:PointMass;
		var node:PointMass;

		p1 = water.pointMasses.tail;
		while(p1.position.x < Clay.camera.pos.x + Clay.screen.width + 32) {
			swapPointMass();
			p1 = water.pointMasses.tail;
		}

		water.sinPos += dt;

		var i:Int = pmOffset;
		node = water.pointMasses.head;
		while(node != null) {
			sineMove(node, i);
			i++;
			node = node.next;
		}

		node = water.pointMasses.head;
		while(node != null) {
			shapeMatching(node, dt);
			node = node.next;
		}

		node = water.pointMasses.head;
		while(node != null) {

			p1 = node;
			p2 = node.next;
			springs(p1, p2, dt);

			node = node.next;

		}

		// Clay.debug.start(broadPhaseName);
		node = water.pointMasses.head;
		while(node != null) {
			broadPhase.updateBroadphase(node);
			node = node.next;
		}
		// Clay.debug.end(broadPhaseName);


		collision();
		

		node = water.pointMasses.head;
		while(node != null) {
			updateVelocity(node, dt);
			integratePosition(node, dt);
			node = node.next;
		}

		surf.line.rotation += surf.angularVelocity * dt;

		// if(playerTransform != null) {
			playerTransform.pos.copy_from(surf.line.position);
		// }
		
		// if(playerSprite != null) {
			playerSprite.rotation_z = surf.line.rotation;
		// }

		updateGeometry();

	}

	function stop():Void {

		if(!stopped) {
			stopped = true;

			Actuate.tween(this, 2, {speed: 0}).ease(Quad.easeOut).onComplete(function() {
				Actuate.tween(playerTransform.scale, 0.5, {x : 0, y : 0}).onComplete(resetPlayer).ease(Quad.easeOut);
				});
		}

	}

	function resetPlayer():Void {

		surf.velocity.y = 0;
		surf.angularVelocity = 0;
		surf.line.rotation = 0;
		surf.line.y = spawnPos;
		playerSprite.rotation_z = 0;
		playerTransform.pos.y = spawnPos;

		Actuate.tween(playerTransform.scale, 0.5, {x : 1, y : 1}).ease(Quad.easeOut).onComplete(function() {

			Actuate.tween(this, 2, {speed: startSpeed}).ease(Quad.easeOut).onComplete(function() {
				stopped = false;
			});

		});


	}

	inline function collision():Void {
		
		var p1:PointMass;
		var p2:PointMass;

		// var pA:PointMass;
		// var pB1:PointMass;
		// var pB2:PointMass;


		var cpm:Array<PointMass> = broadPhase.getLineGridPointMasses(surf.line.p0, surf.line.p1);

		var sign:Int = Maths.sign(surf.line.p1.x - surf.line.p0.x);

		if(cpm != null) {
			var n:Int = 0;
			for (p in cpm) {

				// var len:Float = Clay.screen.height - pA.position.y;

				// collisionDetection(p, surf.pmA.position, surf.pmB.position);
				// ShapeDrawer.drawText(16,32, resA);
				var coll:RayIntersection = collisionDetection(p.position, surf.line.p0, surf.line.p1);
				if(coll != null) {
					var invMass:Float = p.invMass + surf.invMass;
					var pen:Float = coll.distance1 / invMass;

					p.velocity.y += pen * p.invMass;
					// surf.pmA.velocity.y -= pen * (1 - coll.fraction2);
					// surf.pmB.velocity.y -= pen * coll.fraction2;

					p.position.y += pen;
					// surf.line.y -= pen;
						
					surf.velocity.y -= pen * surf.invMass;

					surf.angularVelocity -= surf.invInertia * (coll.distance1 * (coll.fraction2 - 0.5)) * sign;

					if(sign < 0) {
						stop();
						// send loose event
					}

					// trace(coll.fraction2 - 0.5);

					// surf.angularVelocity -= (pen * (coll.fraction2 - 0.5)) * surf.invInertia;
					// surf.angularVelocity -= (pen * (coll.fraction2 - 0.5)) * surf.invInertia;
					// cross : a.x * b.y - a.y * b.x;
					// var aimp:Float = coll.point.x * pen - coll.point.y * 0;
					// surf.angularVelocity += aimp; 

					// angularVelocity += 1.0f / inertia * Cross( contactVector, impulse ); 

					// surf.pmA.position.y -= pen * (1 - coll.fraction2);
					// surf.pmB.position.y -= pen * coll.fraction2;

				}
				// ShapeDrawer.drawText(16 + (256 * n),32, coll);
				// ShapeDrawer.drawCircle(p1.position, 8 + (4 * _c));
				n++;

			}
		}

	}


	var bottomPos:Vector = new Vector();
	function collisionDetection(pA:Vector, pB1:Vector, pB2:Vector):RayIntersection {

		bottomPos.x = pA.x;
		bottomPos.y = Clay.screen.height;

		return lineIntersection(pA, bottomPos, pB1, pB2);
	}

    function lineIntersection( start1:Vector, end1:Vector, start2:Vector, end2:Vector, ?into:RayIntersection ) : RayIntersection {

		var delta1X:Float = end1.x - start1.x;
		var delta1Y:Float = end1.y - start1.y;

		var delta2X:Float = end2.x - start2.x;
		var delta2Y:Float = end2.y - start2.y;

        var diffX:Float = start1.x - start2.x;
        var diffY:Float = start1.y - start2.y;

        var ud:Float = delta2Y * delta1X - delta2X * delta1Y;

        if(ud == 0.0) {
            return null;
        }

        var u1:Float = (delta2X * diffY - delta2Y * diffX) / ud;
        var u2:Float = (delta1X * diffY - delta1Y * diffX) / ud;

        if ((u1 > 0.0 && u1 <= 1) && (u2 > 0.0 && u2 <= 1)) {
            into = (into == null) ? new RayIntersection() : into.reset();

                into.centroid1.x = start1.x;
                into.centroid1.y = start1.y;

                into.centroid2.x = start2.x;
                into.centroid2.y = start2.y;

                into.fraction1 = u1;
                into.fraction2 = u2;

            	var dx:Float = delta1X * u1;
				var dy:Float = delta1Y * u1;

                into.point.x = start1.x + dx;
	        	into.point.y = start1.y + dy;

            	into.distance1 = Math.sqrt(dx * dx + dy * dy);

				dx = delta2X * u2;
				dy = delta2Y * u2;

            	into.distance2 = Math.sqrt(dx * dx + dy * dy);

            return into;
        }

        return null;

    }


#if waterDebug

	override function onRender() {

		if(surf != null) {
			ShapeDrawer.drawLine(surf.line.p0, surf.line.p1, ShapeDrawer.color_pink);
			ShapeDrawer.drawPoint(surf.line.x, surf.line.y, ShapeDrawer.color_red);
			// trace('line.p0: ${line.p0}');
			// trace('line.p1: ${line.p1}');
		}

/*		if(surfNodes != null) {
			
			for (node in surfNodes) {
				ShapeDrawer.drawLine(node.surf.pmA.position, node.surf.pmB.position);
			}
		}*/

		if(water != null) {

			var node:PointMass = water.pointMasses.head;
			while(node != null) {
				drawPointMass(node);
				node = node.next;
			}

			var p1:PointMass;
			var p2:PointMass;
			node = water.pointMasses.head;
			while(node != null) {

				p1 = node;
				p2 = node.next;
				node = node.next;

				if(p2 != null) {
					ShapeDrawer.drawLine(p1.position, p2.position, ShapeDrawer.color_blue);
				}

			}

		}

	}

#end

	override function onMouseMove( e:MouseEvent ) {

		water.spawnOffset = e.pos.y - (Clay.screen.h * 0.5);

		mousePos.copy_from(e.pos);

	}

	override function onKeyUp( e:KeyEvent ) {

		if(e.keycode == Key.key_f) {
			surf.line.rotation -= 1;
		}

		if(e.keycode == Key.key_g) {
			surf.line.rotation += 1;
		}


		if(e.keycode == Key.escape) {
			Clay.shutdown();
		}

		if(e.keycode == Key.key_a) {
			speed -= 8;
			if(speed < 0 ) {
				speed = 0;
			}
		}

		if(e.keycode == Key.key_s) {
			speed += 8;
		}

		if(e.keycode == Key.key_r) {
			if(water != null) {
				// var rnd:Float = 1000;
				// water.pointMasses[Clay.utils.random.int(water.pointMasses.length)].velocity.y += Clay.utils.random.float(-rnd,rnd);
				// water.pointMasses[Clay.utils.random.int(water.pointMasses.length)].velocity.y += rnd;
				// for (p in water.pointMasses) {
				// 	randomVel(p);
				// }
			}
		}

	} //onkeyup

	inline function sineMove(p:PointMass, idx:Int):Void {

		var wave:Float = 2 * Math.sin(idx + water.sinPos * 2);
		wave += 4 * Math.sin(idx * 0.2 + water.sinPos);
		wave += 4 * Math.sin(idx * 0.4 + water.sinPos);
		p.basePosition.y = wave + (Clay.screen.h * 0.5) + p.offset;

		if(p.swapped) {
			p.swapped = false;
			var posAB:Float = p.basePosition.y;
			if(p.prev != null) {
				posAB = (p.prev.position.y + p.basePosition.y) * 0.5;
			}
			p.position.y = posAB;
			p.velocity.y = 0;
		}

	}

	var tmpVec:Vector = new Vector();
	inline function shapeMatching(p:PointMass, dt:Float) {

		tmpVec.x = (p.basePosition.x - p.lastBasePosition.x) / dt; 
		tmpVec.y = (p.basePosition.y - p.lastBasePosition.y) / dt;

		if(water.baseStiffness > 0) {
			var force:Vector = calculateSpringForce(
				p.position, 
				p.velocity,
				p.basePosition,
				tmpVec,
				0, 
				water.baseStiffness, 
				water.baseDamping
			);

			// p.force.x += force.x * p.shapeMatchingMultiplier;
			p.force.y += force.y * p.shapeMatchingMultiplier;

		}

	}

	inline function springs(p1:PointMass, p2:PointMass, dt:Float) {

		if(p2 != null && water.springStiffness > 0) {
			var force:Vector = calculateSpringForce(
				p1.position, 
				p1.velocity,
				p2.position,
				p2.velocity,
				water.distance, 
				water.springStiffness, 
				water.springDamping
			);

			// p.force.x += force.x * p.shapeMatchingMultiplier;
			p1.force.y += force.y * p1.shapeMatchingMultiplier;
			p2.force.y -= force.y * p2.shapeMatchingMultiplier;

		}

	}

	// var rndVel:Float = 512;
	var rndVel:Float = 16;
	inline function randomVel(p:PointMass) {

		if(Clay.utils.random.bool()) {
			// p.velocity.x += Clay.utils.random.float(-rndVel, rndVel);
			p.basePosition.y += Clay.utils.random.float(-rndVel, rndVel);
		}

	}

	inline function updateVelocity(p:PointMass, dt:Float) {

		// p.velocity.y += water.gravity * dt;

		p.velocity.x += (p.force.x * p.invMass) * dt;
		p.velocity.y += (p.force.y * p.invMass) * dt;

		p.force.x = 0;
		p.force.y = 0;
		
	}

	inline function clampVelocity(p:PointMass) {

		p.velocity.x = Maths.clamp(p.velocity.x, -p.maxSpeed, p.maxSpeed);
		p.velocity.y = Maths.clamp(p.velocity.y, -p.maxSpeed, p.maxSpeed);

	}

	inline function integratePosition(p:PointMass, dt:Float) {

		p.lastBasePosition.copy_from(p.basePosition); // ?

		p.position.x += p.velocity.x * dt;
		p.position.y += p.velocity.y * dt;

	}

	inline function drawPointMass(p:PointMass) {
		
		ShapeDrawer.drawCircle(p.basePosition, 2, ShapeDrawer.color_gray_05);
		ShapeDrawer.drawCircle(p.position, 4, ShapeDrawer.color_blue);

	}

	inline function swapPointMass() {

		var p:PointMass = water.pointMasses.remFirst();
		var g:WaterGeometry = water.geometry.pop();

		var pLast:PointMass = water.pointMasses.tail;
		p.offset = water.spawnOffset;
		p.position.x = pLast.position.x + water.distance;
		p.basePosition.x = pLast.basePosition.x + water.distance;
		p.lastBasePosition.x = pLast.lastBasePosition.x + water.distance;
		p.swapped = true;
		broadPhase.updateBroadphase(p);

		g.pA = pLast;
		g.pB = p;

		g.geometry.transform.pos.x = (pLast.position.x + p.position.x) * 0.5 - water.distance * 0.5;

		water.pointMasses.addLast(p);
		water.geometry.add(g);

		pmOffset++;
		if(pmOffset == water.pointMasses.length) {
			pmOffset = 0;
		}

	}


	inline function updateGeometry() {

		for (g in water.geometry) {
			g.geometry.vertices[0].pos.y = g.pA.position.y - 320;
			g.geometry.vertices[4].pos.y = g.pA.position.y - 320;
			g.geometry.vertices[1].pos.y = g.pB.position.y - 320;
		}

	}

	inline function createGeometry() {

		var wg:WaterGeometry;

		var node:PointMass = water.pointMasses.head;

		var p1:PointMass;
		var p2:PointMass;

		while(node != null) {

			p1 = node;
			p2 = node.next;

			node = node.next;

			if(p2 != null) {

				wg = new WaterGeometry();
				wg.pA = p1;
				wg.pB = p2;
				wg.geometry = new QuadGeometry({
					id : 'water geom',
					x : (p1.position.x + p2.position.x) * 0.5 - water.distance * 0.5,
					y : 320,
					w : water.distance,
					h : 320,
					batcher : Clay.renderer.batcher,
					depth : Layers.GROUND
				});

				wg.geometry.vertices[0].color = colorA;
				wg.geometry.vertices[1].color = colorA;
				wg.geometry.vertices[4].color = colorA;
				
				wg.geometry.vertices[2].color = colorB;
				wg.geometry.vertices[3].color = colorB;
				wg.geometry.vertices[5].color = colorB;

				water.geometry.add(wg);

			}

		}

	}

	var _btoaTmp:Vector = new Vector();
	inline function calculateSpringForce(posA:Vector, velA:Vector, posB:Vector, velB:Vector, distance:Float, stiff:Float, damp:Float):Vector {
		
		_btoaTmp.set_xy(0,0);

		if(posA.x == posB.x && posA.y == posB.y){
			return _btoaTmp;
		}

		var dist:Float = Math.sqrt(((posA.x - posB.x) * (posA.x - posB.x)) + ((posA.y - posB.y) * (posA.y - posB.y)));  // Distance
		_btoaTmp.x = (posA.x - posB.x) / dist;
		_btoaTmp.y = (posA.y - posB.y) / dist;

		dist = distance - dist; 

		var vAmVbX:Float = velA.x - velB.x;
		var vAmVbY:Float = velA.y - velB.y;

		var dotAB:Float = vAmVbX * _btoaTmp.x + vAmVbY * _btoaTmp.y;

		_btoaTmp.multiplyScalar((dist * stiff) - dotAB * damp);

		return _btoaTmp;  

	}

	function onWaterNodeAdded(node:WaterNode) {

		water = node.water;

		var len:Int = Math.round((Clay.screen.w + 480 + 32) / water.distance);

		for (i in 0...len) {

			var p:PointMass = new PointMass();
			p.basePosition.set_xy((i * water.distance) - 480, 320);
			p.position.set_xy((i * water.distance) - 480, 320);
			broadPhase.updateBroadphase(p);
			water.pointMasses.addLast(p);

		}

		createGeometry();

	}
	
	function onWaterNodeRemoved(node:WaterNode) {

		water = null;

	}

	function onSurfNodeAdded(node:SurfboardNode) {

		surf = node.surf;

		// surf.pmA = new PointMass();
		// surf.pmA.basePosition.set_xy(160, 128);
		// surf.pmA.position.set_xy(160, 128);

		// surf.pmB = new PointMass();
		// surf.pmB.basePosition.set_xy(280, 256);
		// surf.pmB.position.set_xy(280, 256);

		surf.line = new Line(256,128, 128);
		surf.velocity = new Vector();
	}
	
	function onSurfNodeRemoved(node:SurfboardNode) {

		surfNodes = null;

	}

	function onPlayerNodeAdded(node:PlayerNode) {

		// surf = node.surf;
		// trace('onPlayerNodeAdded');
		// trace(node.sprite);
		// trace(node.transform);
		playerSprite = node.sprite;
		playerSprite.texture = Clay.resources.texture('assets/player.png');

		playerTransform = node.transform;

	}
	
	function onPlayerNodeRemoved(node:PlayerNode) {
		trace('onPlayerNodeRemoved'); // why is removed all time ???

		playerNodes = null;
		playerSprite = null;
		playerTransform = null;

	}


}
